@@ -7,17 +7,18 @@
 
 #include <map>
 #include <memory>  // for unique_ptr
-#include <ostream>
 #include <regex>
 #include "cifdoc.hpp"  // for cif::Document
+#include "logger.hpp"  // for Logger
 
 namespace gemmi { namespace cif {
 
 /// Represents DDL1 or DDL2 dictionary (ontology).
 struct GEMMI_DLL Ddl {
+  /// member functions use logger's callback and threshold for output
+  Logger logger;
   // configuration - some of these flag must be set before read_ddl()
   bool print_unknown_tags = true;
-  bool print_extra_diagnostics = false;
   // these flags below are relevant to DDL2 only
   bool use_regex = true;
   bool use_context = false;
@@ -36,11 +37,15 @@ struct GEMMI_DLL Ddl {
   Ddl(Ddl const&) = delete;
   Ddl& operator=(Ddl const&) = delete;
 
-  void read_ddl(cif::Document&& doc, std::ostream& out);
+  /// it moves doc to ddl_docs_ to control lifetime and prevent modifications
+  void read_ddl(cif::Document&& doc);
 
-  bool validate_cif(const cif::Document& doc, std::ostream& out) const;
+  bool validate_cif(const cif::Document& doc) const;
+  bool validate_block(const cif::Block& b, const std::string& source) const;
 
-  void check_audit_conform(const cif::Document& doc, std::ostream& out) const;
+  void check_audit_conform(const cif::Document& doc) const;
+
+  const std::map<std::string, std::regex>& regexes() const { return regexes_; }
 
 private:
   // items from DDL2 _pdbx_item_linked_group[_list]
@@ -61,12 +66,16 @@ private:
     auto iter = name_index_.find(to_lower(name));
     return iter != name_index_.end() ? iter->second : nullptr;
   }
-  void check_mandatory_items(const cif::Block& b, std::ostream& out) const;
-  void check_unique_keys_in_loop(const cif::Loop& loop, std::ostream& out,
-                                 const std::string& block_name) const;
-  void check_parents(const cif::Block& b, std::ostream& out) const;
+  void check_mandatory_items(const cif::Block& b) const;
+  void check_unique_keys_in_loop(const cif::Loop& loop, const cif::Block& block) const;
+  void check_parents(const cif::Block& b) const;
+  void check_parent_link(const ParentLink& link, const cif::Block& b) const;
   void read_ddl1_block(cif::Block& block);
-  void read_ddl2_block(cif::Block& block, std::ostream& out);
+  void read_ddl2_block(cif::Block& block);
+
+  template<class... Args> void warn(const cif::Block& b, Args const&... args) const {
+    logger.level<3>('[', b.name, "] ", args...);
+  }
 };
 
 }} // namespace gemmi::cif
