@@ -9,7 +9,6 @@
 #include <cstddef>    // for ptrdiff_t
 #include <complex>
 #include <algorithm>  // for fill
-#include <memory>     // for unique_ptr
 #include <numeric>    // for accumulate
 #include <type_traits>
 #include <vector>
@@ -76,7 +75,7 @@ inline int round_with_small_factorization(double exact, GridSizeRounding roundin
   return n;
 }
 
-inline std::array<int, 3> good_grid_size(const std::array<double, 3>& limit,
+inline std::array<int, 3> good_grid_size(std::array<double, 3> limit,
                                          GridSizeRounding rounding,
                                          const SpaceGroup* sg) {
   std::array<int, 3> m = {{0, 0, 0}};
@@ -84,25 +83,33 @@ inline std::array<int, 3> good_grid_size(const std::array<double, 3>& limit,
   if (sg)
     gops = sg->operations();
   std::array<int, 3> sg_fac = gops.find_grid_factors();
-  for (int i = 0; i != 3; ++i) {
+
+  // If two dimensions are symmetry-related, they must have the same size.
+  // Otherwise, if two dimensions are almost equal, the same grid size
+  // looks better and is preferred.
+  for (int i = 1; i < 3; ++i)
     for (int j = 0; j < i; ++j)
-      if (std::fabs(limit[i] - limit[j]) < 0.5 && sg_fac[i] == sg_fac[j]) {
-        m[i] = m[j];
-        break;
+      if (gops.are_directions_symmetry_related(i, j) ||
+          (std::fabs(limit[i] - limit[j]) < 0.5 && sg_fac[i] == sg_fac[j])) {
+        if (rounding == GridSizeRounding::Up)
+          limit[j] = std::max(limit[i], limit[j]);
+        else if (rounding == GridSizeRounding::Down)
+          limit[j] = std::min(limit[i], limit[j]);
+        else // GridSizeRounding::Nearest
+          limit[j] = 0.5 * (limit[i] + limit[j]);
+        sg_fac[i] = -j;  // mark the dimension with higher index
       }
-    if (m[i] == 0) {
-      // having sizes always even simplifies things
-      int f = sg_fac[i] % 2 == 0 ? sg_fac[i] : 2 * sg_fac[i];
-      int n = round_with_small_factorization(limit[i] / f, rounding);
-      m[i] = n * f;
+
+  for (int i = 0; i < 3; ++i) {
+    int f = sg_fac[i];
+    if (f > 0) {  // need to calculate the size
+      if (f % 2 != 0)
+        f *= 2; // always use even sizes - it simplifies things
+      m[i] = f * round_with_small_factorization(limit[i] / f, rounding);
+    } else { // the same size was already calculated
+      m[i] = m[-f];
     }
   }
-  for (int i = 1; i != 3; ++i)
-    for (int j = 0; j != i; ++j)
-      if (gops.are_directions_symmetry_related(i, j) && m[i] != m[j]) {
-        bool denser = rounding != GridSizeRounding::Down;
-        m[i] = m[j] = (denser ? std::max(m[i], m[j]) : std::min(m[i], m[j]));
-      }
   return m;
 }
 
@@ -203,6 +210,9 @@ struct GridMeta {
   size_t index_q(int u, int v, int w) const {
     return size_t(w * nv + v) * nu + u;
   }
+  size_t index_q(size_t u, size_t v, size_t w) const {
+    return (w * nv + v) * nu + u;
+  }
 
   /// Faster than index_s(), but works only if `-nu <= u < 2*nu`, etc.
   size_t index_n(int u, int v, int w) const { return index_n_ref(u, v, w); }
@@ -310,6 +320,8 @@ struct Grid : GridBase<T> {
 
   /// spacing between virtual planes, not between points
   double spacing[3] = {0., 0., 0.};
+  /// unit_cell.orth.mat columns divided by nu, nv, nw
+  UpperTriangularMat33 orth_n;
 
   /// copy unit_cell, spacegroup, nu, nv, nw, axis_order and set spacing
   void copy_metadata_from(const GridMeta& g) {
@@ -322,15 +334,15 @@ struct Grid : GridBase<T> {
     calculate_spacing();
   }
 
-  /// set #spacing
+  /// set #spacing and #orth_n
   void calculate_spacing() {
     spacing[0] = 1.0 / (nu * unit_cell.ar);
     spacing[1] = 1.0 / (nv * unit_cell.br);
     spacing[2] = 1.0 / (nw * unit_cell.cr);
-  }
-
-  double min_spacing() const {
-    return std::min(std::min(spacing[0], spacing[1]), spacing[2]);
+    Vec3 inv_n(1.0 / nu, 1.0 / nv, 1.0 / nw);
+    orth_n = unit_cell.orth.mat.multiply_by_diagonal(inv_n);
+    if (!unit_cell.orth.mat.is_upper_triangular())
+      fail("Grids work only with the standard orientation of crystal frame (SCALEn)");
   }
 
   void set_size_without_checking(int nu_, int nv_, int nw_) {
@@ -344,11 +356,10 @@ struct Grid : GridBase<T> {
     set_size_without_checking(nu_, nv_, nw_);
   }
 
-  // The resulting spacing can be smaller (if denser=true) or greater than arg.
   void set_size_from_spacing(double approx_spacing, GridSizeRounding rounding) {
-    std::array<double, 3> limit = {{1. / (unit_cell.ar * approx_spacing),
-                                    1. / (unit_cell.br * approx_spacing),
-                                    1. / (unit_cell.cr * approx_spacing)}};
+    std::array<double, 3> limit = {{unit_cell.a / approx_spacing,
+                                    unit_cell.b / approx_spacing,
+                                    unit_cell.c / approx_spacing}};
     auto m = good_grid_size(limit, rounding, spacegroup);
     set_size_without_checking(m[0], m[1], m[2]);
   }
@@ -365,10 +376,11 @@ struct Grid : GridBase<T> {
   }
 
   template<typename S>
-  void setup_from(const S& st, double approx_spacing) {
+  void setup_from(const S& st, double approx_spacing=0) {
     spacegroup = st.find_spacegroup();
-    set_unit_cell(st.cell);
-    set_size_from_spacing(approx_spacing, GridSizeRounding::Up);
+    unit_cell = st.cell;
+    if (approx_spacing > 0)
+      set_size_from_spacing(approx_spacing, GridSizeRounding::Up);
   }
 
   /// Returns index in data array for (u,v,w). Safe but slower than index_q().
@@ -423,7 +435,7 @@ struct Grid : GridBase<T> {
 
   /// https://en.wikipedia.org/wiki/Trilinear_interpolation
   /// x,y,z are grid coordinates (x=1.5 is between 2nd and 3rd grid point).
-  T interpolate_value(double x, double y, double z) const {
+  T trilinear_interpolation(double x, double y, double z) const {
     this->check_not_empty();
     int u, v, w;
     double xd = grid_modulo(x, nu, &u);
@@ -444,11 +456,11 @@ struct Grid : GridBase<T> {
     }
     return (T) lerp_(avg[0], avg[1], zd);
   }
-  T interpolate_value(const Fractional& fctr) const {
-    return interpolate_value(fctr.x * nu, fctr.y * nv, fctr.z * nw);
+  T trilinear_interpolation(const Fractional& fctr) const {
+    return trilinear_interpolation(fctr.x * nu, fctr.y * nv, fctr.z * nw);
   }
-  T interpolate_value(const Position& ctr) const {
-    return interpolate_value(unit_cell.fractionalize(ctr));
+  T trilinear_interpolation(const Position& ctr) const {
+    return trilinear_interpolation(unit_cell.fractionalize(ctr));
   }
 
   /// https://en.wikipedia.org/wiki/Tricubic_interpolation
@@ -505,9 +517,9 @@ struct Grid : GridBase<T> {
   void copy_4x4x4(double& x, double& y, double& z,
                   std::array<std::array<std::array<T,4>,4>,4>& copy) const {
     this->check_not_empty();
-    auto prepare_indices = [this](double& r, int nt, int (&indices)[4]) {
+    auto prepare_indices = [](double& r, int nt, int (&indices)[4]) {
       int t;
-      r = this->grid_modulo(r, nt, &t);
+      r = grid_modulo(r, nt, &t);
       indices[0] = (t != 0 ? t : nt) - 1;
       indices[1] = t;
       if (t + 2 < nt) {
@@ -528,14 +540,17 @@ struct Grid : GridBase<T> {
           copy[i][j][k] = this->get_value_q(u_indices[i], v_indices[j], w_indices[k]);
   }
 
-  /// @param order 1=nearest, 2=linear, 3=cubic interpolation
-  T interpolate(const Fractional& f, int order) const {
+  /// @param order 0=nearest, 1=linear, 3=cubic interpolation
+  T interpolate_value(const Fractional& f, int order=1) const {
     switch (order) {
-      case 1: return *const_cast<Grid<T>*>(this)->get_nearest_point(f).value;
-      case 2: return interpolate_value(f);
+      case 0: return *const_cast<Grid<T>*>(this)->get_nearest_point(f).value;
+      case 1: return trilinear_interpolation(f);
       case 3: return (T) tricubic_interpolation(f);
     }
-    throw std::invalid_argument("interpolation \"order\" must 1, 2 or 3");
+    throw std::invalid_argument("interpolation \"order\" must 0, 1 or 3");
+  }
+  T interpolate_value(const Position& ctr, int order=1) const {
+    return interpolate_value(unit_cell.fractionalize(ctr), order);
   }
 
   void get_subarray(T* dest, std::array<int,3> start, std::array<int,3> shape) const {
@@ -608,10 +623,13 @@ struct Grid : GridBase<T> {
   }
 
   template <bool UsePbc, typename Func>
-  void do_use_points_in_box(Fractional fctr, int du, int dv, int dw, Func&& func) {
-    int u0 = iround(fctr.x * nu);
-    int v0 = iround(fctr.y * nv);
-    int w0 = iround(fctr.z * nw);
+  void do_use_points_in_box(const Fractional& fctr, int du, int dv, int dw, Func&& func,
+                            double radius=INFINITY) {
+    double max_dist_sq = radius * radius;
+    const Fractional nctr(fctr.x * nu, fctr.y * nv, fctr.z * nw);
+    int u0 = iround(nctr.x);
+    int v0 = iround(nctr.y);
+    int w0 = iround(nctr.z);
     int u_lo = u0 - du;
     int u_hi = u0 + du;
     int v_lo = v0 - dv;
@@ -626,45 +644,58 @@ struct Grid : GridBase<T> {
       w_lo = std::max(w_lo, 0);
       w_hi = std::min(w_hi, nw - 1);
     }
-    const Position orth0(unit_cell.orth.mat.column_copy(0));
-    for (int w = w_lo; w <= w_hi; ++w) {
-      int w_ = UsePbc ? modulo(w, nw) : w;
-      double fw = w * (1.0 / nw);
-      for (int v = v_lo; v <= v_hi; ++v) {
-        int v_ = UsePbc ? modulo(v, nv) : v;
-        double fv = v * (1.0 / nv);
-        size_t idx0 = this->index_q(0, v_, w_);
-        Position delta0 = unit_cell.orthogonalize_difference(fctr - Fractional(0., fv, fw));
-        for (int u = u_lo; u <= u_hi; ++u) {
-          int u_ = UsePbc ? modulo(u, nu) : u;
-          double fu = u * (1.0 / nu);
-          Position delta = delta0 - orth0 * fu;
-          func(data[idx0 + u_], delta, u, v, w);
+    int u_0 = UsePbc ? modulo(u_lo, nu) : u_lo;
+    int v_0 = UsePbc ? modulo(v_lo, nv) : v_lo;
+    int w_0 = UsePbc ? modulo(w_lo, nw) : w_lo;
+    auto wrap = [](int& q, int nq) { if (UsePbc && q == nq) q = 0; };
+    Fractional fdelta(nctr.x - u_lo, 0, 0);
+    for (int w = w_lo, w_ = w_0; w <= w_hi; ++w, wrap(++w_, nw)) {
+      fdelta.z = nctr.z - w;
+      for (int v = v_lo, v_ = v_0; v <= v_hi; ++v, wrap(++v_, nv)) {
+        fdelta.y = nctr.y - v;
+        Position delta(orth_n.multiply(fdelta));
+        T* t = &data[this->index_q(u_0, v_, w_)];
+        double dist_sq0 = sq(delta.y) + sq(delta.z);
+        if (dist_sq0 > max_dist_sq)
+          continue;
+        for (int u = u_lo, u_ = u_0;;) {
+          double dist_sq = dist_sq0 + sq(delta.x);
+          if (!(dist_sq > max_dist_sq))
+            func(*t, dist_sq, delta, u, v, w);
+          if (u >= u_hi)
+            break;
+          ++u;
+          ++u_;
+          ++t;
+          if (UsePbc && u_ == nu) {
+            u_ = 0;
+            t -= nu;
+          }
+          delta.x -= orth_n.a11;
         }
       }
     }
   }
 
   template <bool UsePbc, typename Func>
-  void use_points_in_box(Fractional fctr, int du, int dv, int dw,
-                         Func&& func, bool fail_on_too_large_radius=true) {
+  void use_points_in_box(const Fractional& fctr, int du, int dv, int dw,
+                         Func&& func, bool fail_on_too_large_radius=true,
+                         double radius=INFINITY) {
     check_size_for_points_in_box<UsePbc>(du, dv, dw, fail_on_too_large_radius);
-    do_use_points_in_box<UsePbc>(fctr, du, dv, dw, func);
+    do_use_points_in_box<UsePbc>(fctr, du, dv, dw, func, radius);
   }
 
   template <bool UsePbc, typename Func>
-  void use_points_around(const Fractional& fctr_, double radius, Func&& func,
+  void use_points_around(const Fractional& fctr, double radius, Func&& func,
                          bool fail_on_too_large_radius=true) {
     int du = (int) std::ceil(radius / spacing[0]);
     int dv = (int) std::ceil(radius / spacing[1]);
     int dw = (int) std::ceil(radius / spacing[2]);
-    use_points_in_box<UsePbc>(fctr_, du, dv, dw,
-                      [&](T& ref, const Position& delta, int, int, int) {
-                        double d2 = delta.length_sq();
-                        if (d2 < radius * radius)
-                          func(ref, d2);
-                      },
-                      fail_on_too_large_radius);
+    use_points_in_box<UsePbc>(
+        fctr, du, dv, dw,
+        [&](T& ref, double d2, const Position&, int, int, int) { func(ref, d2); },
+        fail_on_too_large_radius,
+        radius);
   }
 
   void set_points_around(const Position& ctr, double radius, T value, bool use_pbc=true) {
@@ -675,6 +706,7 @@ struct Grid : GridBase<T> {
       use_points_around<false>(fctr, radius, [&](T& ref, double) { ref = value; }, false);
   }
 
+  /// Change all occurrences of old_value to new_value
   void change_values(T old_value, T new_value) {
     for (auto& d : data)
       if (impl::is_same(d, old_value))
@@ -694,7 +726,7 @@ struct Grid : GridBase<T> {
     if (ops.empty())
       return;
     std::vector<size_t> mates(ops.size(), 0);
-    std::vector<bool> visited(data.size(), false);
+    std::vector<signed char> visited(data.size(), 0);  // faster than vector<bool>
     size_t idx = 0;
     for (int w = 0; w != nw; ++w)
       for (int v = 0; v != nv; ++v)
@@ -713,10 +745,10 @@ struct Grid : GridBase<T> {
             value = func(value, data[k]);
           }
           data[idx] = value;
-          visited[idx] = true;
+          visited[idx] = 1;
           for (size_t k : mates) {
             data[k] = value;
-            visited[k] = true;
+            visited[k] = 1;
           }
         }
     assert(idx == data.size());
@@ -739,6 +771,14 @@ struct Grid : GridBase<T> {
   void symmetrize_nondefault(T default_) {
     symmetrize([default_](T a, T b) { return impl::is_same(a, default_) ? b : a; });
   }
+  void symmetrize_avg() {
+    symmetrize_sum();
+    if (spacegroup && spacegroup->number != 1) {
+      int n_ops = spacegroup->operations().order();
+      for (T& x : data)
+        x /= n_ops;
+    }
+  }
 
   /// scale the data to get mean == 0 and rmsd == 1 (doesn't work for T=complex)
   void normalize() {
@@ -746,20 +786,23 @@ struct Grid : GridBase<T> {
     for (T& x : data)
       x = static_cast<T>((x - stats.dmean) / stats.rms);
   }
-
-  // TODO: can it be replaced with interpolate_grid(dest, src, Transform(), order)?
-  void resample_to(Grid<T>& dest, int order) const {
-    dest.check_not_empty();
-    int idx = 0;
-    for (int w = 0; w < dest.nw; ++w)
-      for (int v = 0; v < dest.nv; ++v)
-        for (int u = 0; u < dest.nu; ++u, ++idx) {
-          const Fractional f = dest.get_fractional(u, v, w);
-          dest.data[idx] = interpolate(f, order);
-        }
-  }
 };
 
+// TODO: add argument Box<Fractional> src_extent
+// cf. interpolate_grid_of_aligned_model2() in solmask.hpp
+// cf interpolate_values in python/grid.cpp
+template<typename T>
+void interpolate_grid(Grid<T>& dest, const Grid<T>& src, const Transform& tr, int order=1) {
+  FTransform frac_tr = src.unit_cell.frac.combine(tr).combine(dest.unit_cell.orth);
+  size_t idx = 0;
+  for (int w = 0; w != dest.nw; ++w)
+    for (int v = 0; v != dest.nv; ++v)
+      for (int u = 0; u != dest.nu; ++u, ++idx) {
+        Fractional dest_fr = dest.get_fractional(u, v, w);
+        Fractional src_fr = frac_tr.apply(dest_fr);
+        dest.data[idx] = src.interpolate_value(src_fr, order);
+      }
+}
 
 template<typename T>
 Correlation calculate_correlation(const GridBase<T>& a, const GridBase<T>& b) {
