@@ -1,6 +1,6 @@
 // Copyright 2017 Global Phasing Ltd.
 //
-// Data structures to keep macromolecular structure model.
+// Data structures to store macromolecular structure models.
 
 #ifndef GEMMI_MODEL_HPP_
 #define GEMMI_MODEL_HPP_
@@ -29,31 +29,40 @@ namespace gemmi {
 namespace impl {
 
 template<typename T>
-T* find_or_null(std::vector<T>& vec, const std::string& name) {
-  auto it = std::find_if(vec.begin(), vec.end(),
-                         [&name](const T& m) { return m.name == name; });
+auto get_id(const T& m) -> decltype(m.name) { return m.name; }
+template<typename T>
+auto get_id(const T& m) -> decltype(m.num) { return m.num; }
+
+
+template<typename Vec, typename S>
+auto find_iter_(Vec& vec, const S& name) {
+  return std::find_if(vec.begin(), vec.end(), [&name](const auto& m) { return get_id(m) == name; });
+}
+
+template<typename T, typename S>
+T* find_or_null(std::vector<T>& vec, const S& name) {
+  auto it = find_iter_(vec, name);
   return it != vec.end() ? &*it : nullptr;
 }
 
-template<typename T>
-T& find_or_add(std::vector<T>& vec, const std::string& name) {
+template<typename T, typename S>
+T& find_or_add(std::vector<T>& vec, const S& name) {
   if (T* ret = find_or_null(vec, name))
     return *ret;
   vec.emplace_back(name);
   return vec.back();
 }
 
-template<typename Span, typename T = typename Span::value_type>
-typename Span::iterator find_iter(Span& span, const std::string& name) {
-  auto i = std::find_if(span.begin(), span.end(),
-                        [&](const T& x) { return x.name == name; });
-  if (i == span.end())
-    throw std::invalid_argument(
-        T::what() + (" " + name) + " not found (only [" +
-        join_str(span.begin(), span.end(), ' ',
-                 [](const T& x) { return x.name; }) +
-        "])");
-  return i;
+template<typename Span, typename S>
+typename Span::iterator find_iter(Span& span, const S& name) {
+  using T = typename Span::value_type;
+  auto it = find_iter_(span, name);
+  if (it == span.end())
+    throw std::invalid_argument(cat(
+        T::what(), ' ', name, " not found (only [",
+        join_str(span.begin(), span.end(), ' ', [](const T& item) { return cat(get_id(item)); }),
+        "])"));
+  return it;
 }
 
 template<typename Group>
@@ -76,7 +85,9 @@ template<typename T, typename M> std::vector<T> model_subchains(M* model) {
 /// File format of a macromolecular model. When passed to read_structure():
 /// Unknown = guess format from the extension,
 /// Detect = guess format from the content.
-enum class CoorFormat { Unknown, Detect, Pdb, Mmcif, Mmjson, ChemComp };
+enum class CoorFormat : unsigned char {
+  Unknown, Detect, Pdb, Mmcif, Mmjson, ChemComp
+};
 
 /// corresponds to _atom_site.calc_flag in mmCIF
 enum class CalcFlag : signed char {
@@ -86,9 +97,6 @@ enum class CalcFlag : signed char {
   NotSet=0, NoHydrogen, Determined, Calculated, Dummy
 };
 
-/// helper type used for Structure::shortened_ccd_codes
-struct OldToNew { std::string old, new_; };
-
 /// options affecting how pdb file is read
 struct PdbReadOptions {
   int max_line_length = 0;
@@ -193,10 +201,12 @@ struct Residue : public ResidueId {
   Residue empty_copy() const {
     Residue res((ResidueId&)*this);
     res.subchain = subchain;
+    res.entity_id = entity_id;
     res.label_seq = label_seq;
     res.entity_type = entity_type;
     res.het_flag = het_flag;
     res.flag = flag;
+    res.sifts_unp = sifts_unp;
     return res;
   }
   using child_type = Atom;
@@ -210,15 +220,20 @@ struct Residue : public ResidueId {
     return nullptr;
   }
 
-  // default values accept anything
-  Atom* find_atom(const std::string& atom_name, char altloc, El el=El::X) {
+  /// El::X means anything;
+  /// in strict_altloc mode, '*' = any altloc, otherwise it's \0
+  Atom* find_atom(const std::string& atom_name, char altloc, El el=El::X,
+                  bool strict_altloc=true) {
+    if (!strict_altloc && altloc == '\0')
+      altloc = '*';
     for (Atom& a : atoms)
       if (a.name == atom_name && a.altloc_matches(altloc) && (el == El::X || a.element == el))
         return &a;
     return nullptr;
   }
-  const Atom* find_atom(const std::string& atom_name, char altloc, El el=El::X) const {
-    return const_cast<Residue*>(this)->find_atom(atom_name, altloc, el);
+  const Atom* find_atom(const std::string& atom_name, char altloc, El el=El::X,
+                        bool strict_altloc=true) const {
+    return const_cast<Residue*>(this)->find_atom(atom_name, altloc, el, strict_altloc);
   }
 
   std::vector<Atom>::iterator find_atom_iter(const std::string& atom_name,
@@ -273,6 +288,12 @@ struct Residue : public ResidueId {
   ConstUniqProxy<Atom> first_conformer() const { return {atoms}; }
 };
 
+inline void add_distinct_altlocs(const Residue& res, std::string& altlocs) {
+  for (const Atom& atom : res.atoms)
+    if (atom.altloc && altlocs.find(atom.altloc) == std::string::npos)
+      altlocs += atom.altloc;
+}
+
 struct ResidueGroup;
 struct ConstResidueGroup;
 
@@ -325,9 +346,9 @@ struct ConstResidueSpan : Span<const Residue> {
   // We assume residues are ordered. It works (approximately) also with
   // missing numbers which can be present in DBREF.
   SeqId label_seq_id_to_auth(SeqId::OptionalNum label_seq_id) const {
-    if (size() == 0)
+    if (empty())
       throw std::out_of_range("label_seq_id_to_auth(): empty span");
-    auto it = std::lower_bound(begin(), end(), label_seq_id,
+    const auto* it = std::lower_bound(begin(), end(), label_seq_id,
         [](const Residue& r, SeqId::OptionalNum v){ return r.label_seq < v; });
     if (it == end())
       --it;
@@ -341,7 +362,7 @@ struct ConstResidueSpan : Span<const Residue> {
   // The residue numbers (auth) are sometimes not ordered.
   // That is why we use this multi-step heuristic.
   SeqId::OptionalNum auth_seq_id_to_label(SeqId auth_seq_id) const {
-    if (size() == 0)
+    if (empty())
       throw std::out_of_range("auth_seq_id_to_label(): empty span");
     for (const Residue& r : *this)
       if (r.seqid == auth_seq_id)
@@ -383,6 +404,7 @@ struct ResidueSpan : MutableVectorSpan<Residue> {
   GroupingProxy residue_groups();
   const std::string& subchain_id() const { return const_().subchain_id(); }
   ResidueGroup find_residue_group(SeqId id);
+  std::vector<std::string> extract_sequence() const { return const_().extract_sequence(); }
   ConstResidueGroup find_residue_group(SeqId id) const;
   SeqId label_seq_id_to_auth(SeqId::OptionalNum label_seq_id) const {
     return const_().label_seq_id_to_auth(label_seq_id);
@@ -462,7 +484,8 @@ struct Chain {
   std::string name;
   std::vector<Residue> residues;
 
-  explicit Chain(std::string cname) noexcept : name(cname) {}
+  Chain() = default;
+  explicit Chain(const std::string& name_) noexcept : name(name_) {}
 
   ResidueSpan whole() {
     Residue* begin = residues.empty() ? nullptr : &residues[0];
@@ -613,9 +636,10 @@ inline std::string atom_str(const const_CRA& cra) {
                   cra.atom ? cra.atom->altloc : '\0');
 }
 
-inline bool atom_matches(const const_CRA& cra, const AtomAddress& addr) {
+inline bool atom_matches(const const_CRA& cra, const AtomAddress& addr, bool ignore_segment=false) {
   return cra.chain && cra.chain->name == addr.chain_name &&
-         cra.residue && cra.residue->matches(addr.res_id) &&
+         cra.residue && cra.residue->matches_noseg(addr.res_id) &&
+         (ignore_segment || cra.residue->segment == addr.res_id.segment) &&
          cra.atom && cra.atom->name == addr.atom_name &&
          cra.atom->altloc == addr.altloc;
 }
@@ -693,9 +717,11 @@ using ConstCraProxy = CraProxy_<const_CRA, const std::vector<Chain>&>;
 
 struct Model {
   static const char* what() { return "Model"; }
-  std::string name;  // actually an integer number
+  int num = 0;
   std::vector<Chain> chains;
-  explicit Model(std::string mname) noexcept : name(mname) {}
+
+  Model() = default;
+  explicit Model(int num_) noexcept : num(num_) {}
 
   // Returns the first chain with given name, or nullptr.
   Chain* find_chain(const std::string& chain_name) {
@@ -835,7 +861,7 @@ struct Model {
   }
 
   // methods present in Structure, Model, ... - used in templates
-  Model empty_copy() const { return Model(name); }
+  Model empty_copy() const { return Model(num); }
   using child_type = Chain;
   std::vector<Chain>& children() { return chains; }
   const std::vector<Chain>& children() const { return chains; }
@@ -855,6 +881,7 @@ inline const Entity* find_entity_of_subchain(const std::string& subchain_id,
 }
 
 struct Structure {
+  static const char* what() { return "Structure"; }
   std::string name;
   UnitCell cell;
   std::string spacegroup_hm;  // usually pdb symbol, cf. SpaceGroup::pdb_name()
@@ -872,6 +899,8 @@ struct Structure {
 
   CoorFormat input_format = CoorFormat::Unknown;
   bool has_d_fraction = false;  // uses Refmac's ccp4_deuterium_fraction
+  /// in input PDB file: y = TER records were read, e = errors were detected
+  char ter_status = '\0';
 
   /// Store ORIGXn / _database_PDB_matrix.origx*
   bool has_origx = false;
@@ -880,7 +909,7 @@ struct Structure {
   /// Minimal metadata with keys being mmcif tags: _entry.id, _cell.Z_PDB, ...
   std::map<std::string, std::string> info;
   /// Mapping of long (4+) CCD codes (residue names) to PDB-compatible ones
-  std::vector<OldToNew> shortened_ccd_codes;
+  std::vector<std::pair<std::string,std::string>> shortened_ccd_codes;
   /// original REMARK records stored if the file was read from the PDB format
   std::vector<std::string> raw_remarks;
   /// simplistic resolution value from/for REMARK 2
@@ -905,23 +934,19 @@ struct Structure {
     return const_cast<Structure*>(this)->first_model();
   }
 
-  Model* find_model(const std::string& model_name) {
-    return impl::find_or_null(models, model_name);
+  Model* find_model(int model_num) {
+    return impl::find_or_null(models, model_num);
   }
-  const Model* find_model(const std::string& model_name) const {
-    return const_cast<Structure*>(this)->find_model(model_name);
+  const Model* find_model(int model_num) const {
+    return const_cast<Structure*>(this)->find_model(model_num);
   }
-  Model& find_or_add_model(const std::string& model_name) {
-    return impl::find_or_add(models, model_name);
-  }
-
-  void remove_model(const std::string& model_name) {
-    models.erase(impl::find_iter(models, model_name));
+  Model& find_or_add_model(int model_num) {
+    return impl::find_or_add(models, model_num);
   }
 
   void renumber_models() {
     for (size_t i = 0; i != models.size(); ++i)
-      models[i].name = std::to_string(i+1);
+      models[i].num = i + 1;
   }
 
   Entity* get_entity(const std::string& ent_id) {
@@ -950,10 +975,13 @@ struct Structure {
   }
 
   Connection* find_connection_by_cra(const const_CRA& cra1,
-                                     const const_CRA& cra2) {
+                                     const const_CRA& cra2,
+                                     bool ignore_segment=false) {
     for (Connection& c : connections)
-      if ((atom_matches(cra1, c.partner1) && atom_matches(cra2, c.partner2)) ||
-          (atom_matches(cra1, c.partner2) && atom_matches(cra2, c.partner1)))
+      if ((atom_matches(cra1, c.partner1, ignore_segment) &&
+           atom_matches(cra2, c.partner2, ignore_segment)) ||
+          (atom_matches(cra1, c.partner2, ignore_segment) &&
+           atom_matches(cra2, c.partner1, ignore_segment)))
         return &c;
     return nullptr;
   }
@@ -976,10 +1004,10 @@ struct Structure {
     return std::any_of(ncs.begin(), ncs.end(), [](const NcsOp& o) { return !o.given; });
   }
 
-  void add_conect_one_way(int serial1, int serial2, int order) {
-    auto& vec = conect_map[serial1];
+  void add_conect_one_way(int serial_a, int serial_b, int order) {
+    auto& vec = conect_map[serial_a];
     for (int i = 0; i < order; ++i)
-      vec.insert(std::upper_bound(vec.begin(), vec.end(), serial2), serial2);
+      vec.insert(std::upper_bound(vec.begin(), vec.end(), serial_b), serial_b);
   }
   void add_conect(int serial1, int serial2, int order) {
     add_conect_one_way(serial1, serial2, order);
@@ -1005,18 +1033,21 @@ struct Structure {
     st.ncs = ncs;
     st.entities = entities;
     st.connections = connections;
+    st.cispeps = cispeps;
+    st.mod_residues = mod_residues;
     st.helices = helices;
     st.sheets = sheets;
     st.assemblies = assemblies;
     st.meta = meta;
+    st.input_format = input_format;
     st.has_origx = has_origx;
     st.origx = origx;
     st.info = info;
     st.raw_remarks = raw_remarks;
     st.resolution = resolution;
-    st.input_format = input_format;
     return st;
   }
+
   using child_type = Model;
   std::vector<Model>& children() { return models; }
   const std::vector<Model>& children() const { return models; }
