@@ -14,6 +14,8 @@
 #include "mtz.hpp"       // for Mtz
 #include "xds_ascii.hpp" // for XdsAscii
 #include "intensit.hpp"  // for Intensities
+#include "logger.hpp"    // for Logger
+#include "metadata.hpp"  // for SoftwareItem
 
 namespace gemmi {
 
@@ -90,7 +92,7 @@ public:
 
   void write_cif(const Mtz& mtz, const Mtz* mtz2,
                  SMat33<double>* staraniso_b, std::ostream& os);
-  void write_cif_from_xds(const XdsAscii& xds, std::ostream& os);
+  void write_cif_from_xds(const XdsAscii& xds, std::ostream& os) const;
 };
 
 GEMMI_DLL void write_staraniso_b_in_mmcif(const SMat33<double>& b,
@@ -98,13 +100,16 @@ GEMMI_DLL void write_staraniso_b_in_mmcif(const SMat33<double>& b,
                                           char* buf, std::ostream& os);
 
 /// remove '_dataset_name' that can be appended to column names in ccp4i
-GEMMI_DLL void remove_appendix_from_column_names(Mtz& mtz, std::ostream& out);
+GEMMI_DLL void remove_appendix_from_column_names(Mtz& mtz, const Logger& logger);
 
-GEMMI_DLL bool validate_merged_mtz_deposition_columns(const Mtz& mtz, std::ostream& out);
+GEMMI_DLL bool validate_merged_mtz_deposition_columns(const Mtz& mtz, const Logger& logger);
 
 // note: both mi and ui get modified
 GEMMI_DLL bool validate_merged_intensities(Intensities& mi, Intensities& ui,
-                                           bool relaxed_check, std::ostream& out);
+                                           bool relaxed_check, const Logger& logger);
+
+GEMMI_DLL std::vector<SoftwareItem>
+get_software_from_mtz_history(const std::vector<std::string>& history);
 
 } // namespace gemmi
 #endif
