@@ -15,8 +15,6 @@
 
 namespace gemmi {
 
-enum class DataType { Unknown, Unmerged, Mean, Anomalous };
-
 struct ReflnBlock {
   cif::Block block;
   std::string entry_id;
@@ -47,6 +45,8 @@ struct ReflnBlock {
   }
   ReflnBlock& operator=(ReflnBlock&&) = default;
   ReflnBlock& operator=(const ReflnBlock& o) {
+    if (this == &o)
+      return *this;
     block = o.block;
     entry_id = o.entry_id;
     cell = o.cell;
@@ -65,11 +65,13 @@ struct ReflnBlock {
   void check_ok() const { if (!ok()) fail("Invalid ReflnBlock"); }
 
   // position after "_refln." or "_diffrn_refln."
-  size_t tag_offset() const { return refln_loop ? 7 : 14; }
+  size_t tag_offset() const { return default_loop == refln_loop ? 7 : 14; }
 
   void use_unmerged(bool unmerged) {
     default_loop = unmerged ? diffrn_refln_loop : refln_loop;
   }
+  bool is_merged() const { return ok() && default_loop == refln_loop; }
+  // deprecated
   bool is_unmerged() const { return ok() && default_loop == diffrn_refln_loop; }
 
   std::vector<std::string> column_labels() const {
@@ -92,8 +94,12 @@ struct ReflnBlock {
 
   size_t get_column_index(const std::string& tag) const {
     int idx = find_column_index(tag);
-    if (idx == -1)
-      fail("Column not found: " + tag);
+    if (idx == -1) {
+      if (!ok())
+        fail("No reflection data in block ", block.name);
+      const char* category = default_loop == refln_loop ? "_refln." : "_diffrn_refln.";
+      fail("Column not found in block ", block.name, ": ", category, tag);
+    }
     return idx;
   }
 
@@ -175,6 +181,7 @@ inline ReflnBlock get_refln_block(std::vector<cif::Block>&& blocks,
                                   const std::vector<std::string>& labels,
                                   const char* block_name=nullptr) {
   const SpaceGroup* first_sg = nullptr;
+  bool has_block = false;
   for (cif::Block& block : blocks) {
     if (!first_sg)
       if (const std::string* hm = impl::find_spacegroup_hm_value(block)) {
@@ -182,22 +189,24 @@ inline ReflnBlock get_refln_block(std::vector<cif::Block>&& blocks,
         if (first_sg && first_sg->ext == 'H') {
           UnitCell cell;
           impl::set_cell_from_mmcif(block, cell);
-          first_sg = find_spacegroup_by_name(cif::as_string(*hm),
-                                             cell.alpha, cell.gamma);
+          first_sg = find_spacegroup_by_name(cif::as_string(*hm), cell.alpha, cell.gamma);
         }
       }
-    if (block_name && block.name != block_name)
-      continue;
-    if (cif::Loop* loop = block.find_loop("_refln.index_h").get_loop())
-      if (std::all_of(labels.begin(), labels.end(),
-            [&](const std::string& s) { return loop->has_tag("_refln."+s); })) {
-        ReflnBlock rblock(std::move(block));
-        if (!rblock.spacegroup && first_sg)
-          rblock.spacegroup = first_sg;
-        return rblock;
-      }
+    if (!block_name || block.name == block_name) {
+      has_block = true;
+      if (cif::Loop* loop = block.find_loop("_refln.index_h").get_loop())
+        if (std::all_of(labels.begin(), labels.end(),
+              [&](const std::string& s) { return loop->has_tag("_refln."+s); })) {
+          ReflnBlock rblock(std::move(block));
+          if (!rblock.spacegroup && first_sg)
+            rblock.spacegroup = first_sg;
+          return rblock;
+        }
+    }
   }
-  fail("Required block or tags not found in the SF-mmCIF file.");
+  if (block_name && !has_block)
+    fail("Required block not found in SF-mmCIF file.");
+  fail("Tags not found in SF-mmCIF file: _refln.", join_str(labels, ", _refln."));
 }
 
 inline ReflnBlock hkl_cif_as_refln_block(cif::Block& block) {
