@@ -46,7 +46,7 @@ inline float cctbx_vdw_radius(El el) {
     /*Ts*/ 1.00f, /*Og*/ 1.00f,
     /*D*/  1.20f, /*END*/0.f
   };
-  static_assert(radii[static_cast<int>(El::D)] == 1.2f, "Hmm");
+  static_assert(ce_almost_eq(radii[static_cast<int>(El::D)], 1.2f), "Hmm");
   static_assert(sizeof(radii) / sizeof(radii[0]) ==
                 static_cast<int>(El::END) + 1, "Hmm");
   return radii[static_cast<int>(el)];
@@ -87,7 +87,7 @@ inline float refmac_radius_for_bulk_solvent(El el) {
     /*Ts*/ 1.00f, /*Og*/ 1.00f,
     /*D*/  1.40f, /*END*/0.f
   };
-  static_assert(radii[static_cast<int>(El::D)] == 1.40f, "Hmm");
+  static_assert(ce_almost_eq(radii[static_cast<int>(El::D)], 1.40f), "Hmm");
   return radii[static_cast<int>(el)];
 #else
   // temporary solution used in Refmac
@@ -103,27 +103,13 @@ inline float refmac_radius_for_bulk_solvent(El el) {
 }
 
 // mask utilities
-template<typename T>
-void mask_points_in_constant_radius(Grid<T>& mask, const Model& model,
-                                    double radius, T value,
-                                    bool ignore_hydrogen,
-                                    bool ignore_zero_occupancy_atoms) {
-  for (const Chain& chain : model.chains)
-    for (const Residue& res : chain.residues)
-      for (const Atom& atom : res.atoms) {
-        if ((ignore_hydrogen && atom.is_hydrogen()) ||
-            (ignore_zero_occupancy_atoms && atom.occ <= 0))
-          continue;
-        mask.set_points_around(atom.pos, radius, value);
-      }
-}
 
 template<typename T>
-void mask_points_in_varied_radius(Grid<T>& mask, const Model& model,
-                                  AtomicRadiiSet atomic_radii_set,
-                                  double r_probe, T value,
-                                  bool ignore_hydrogen,
-                                  bool ignore_zero_occupancy_atoms) {
+void mask_points_in_radius(Grid<T>& mask, const Model& model,
+                           AtomicRadiiSet atomic_radii_set,
+                           double r_probe, T value,
+                           bool ignore_hydrogen,
+                           bool ignore_zero_occupancy_atoms) {
   for (const Chain& chain : model.chains)
     for (const Residue& res : chain.residues)
       for (const Atom& atom : res.atoms) {
@@ -131,18 +117,101 @@ void mask_points_in_varied_radius(Grid<T>& mask, const Model& model,
             (ignore_zero_occupancy_atoms && atom.occ <= 0))
           continue;
         El elem = atom.element.elem;
-        double r = 0;
+        double r = r_probe;
         switch (atomic_radii_set) {
-          case AtomicRadiiSet::VanDerWaals: r = vdw_radius(elem); break;
-          case AtomicRadiiSet::Cctbx: r = cctbx_vdw_radius(elem); break;
-          case AtomicRadiiSet::Refmac: r = refmac_radius_for_bulk_solvent(elem); break;
-          case AtomicRadiiSet::Constant: assert(0); break;
+          case AtomicRadiiSet::VanDerWaals: r += vdw_radius(elem); break;
+          case AtomicRadiiSet::Cctbx: r += cctbx_vdw_radius(elem); break;
+          case AtomicRadiiSet::Refmac: r += refmac_radius_for_bulk_solvent(elem); break;
+          case AtomicRadiiSet::Constant: /* r is included in r_probe */ break;
         }
-        mask.set_points_around(atom.pos, r + r_probe, value);
+        mask.set_points_around(atom.pos, r, value);
       }
 }
 
-// All points != value in a distance < r from value are set to margin_value
+// deprecated
+template<typename T>
+void mask_points_in_constant_radius(Grid<T>& mask, const Model& model,
+                                    double radius, T value,
+                                    bool ignore_hydrogen,
+                                    bool ignore_zero_occupancy_atoms) {
+  mask_points_in_radius(mask, model, AtomicRadiiSet::Constant, radius, value,
+                        ignore_hydrogen, ignore_zero_occupancy_atoms);
+}
+
+
+inline std::string distinct_altlocs(const Model& model) {
+  std::string altlocs;
+  for (const Chain& chain : model.chains)
+    for (const Residue& res : chain.residues)
+      add_distinct_altlocs(res, altlocs);
+  return altlocs;
+}
+
+/// mask all grid points within a varying radius around model atoms
+/// @param mask existing mask object
+/// @param model model to consider
+/// @param atomic_radii_set set of atomic radii
+/// @param r_probe probe radius to add to atomic radius (in A)
+/// @param value value to use (unless use_atom_occupancy is set to true)
+/// @param ignore_hydrogen should we skip hydrogen atoms
+/// @param ignore_zero_occupancy_atoms should we skip atoms with occupancy of zero
+/// @param use_atom_occupancy should we use the occupancy of each atom (to build up a non-binary mask)
+inline
+void mask_points_using_occupancy(Grid<float>& mask, const Model& model,
+                                  AtomicRadiiSet atomic_radii_set, double r_probe,
+                                  bool ignore_hydrogen,
+                                  bool ignore_zero_occupancy_atoms) {
+
+  std::string altlocs = distinct_altlocs(model);
+  altlocs += '\0';  // no altloc
+
+  gemmi::Grid<float> m;
+  m.copy_metadata_from(mask);
+  for (char& altloc : altlocs) {
+    m.fill(0.0);
+    for (const Chain& chain : model.chains) {
+      for (const Residue& res : chain.residues) {
+        for (const Atom& atom : res.atoms) {
+          if ((ignore_hydrogen && atom.is_hydrogen()) ||
+              (ignore_zero_occupancy_atoms && atom.occ <= 0))
+            continue;
+          if (atom.altloc == altloc) {
+            El elem = atom.element.elem;
+            double r = r_probe;
+            switch (atomic_radii_set) {
+              case AtomicRadiiSet::VanDerWaals: r += vdw_radius(elem); break;
+              case AtomicRadiiSet::Cctbx: r += cctbx_vdw_radius(elem); break;
+              case AtomicRadiiSet::Refmac: r += refmac_radius_for_bulk_solvent(elem); break;
+              case AtomicRadiiSet::Constant: /* r is included in r_probe */ break;
+            }
+            Fractional fpos = m.unit_cell.fractionalize(atom.pos);
+            m.use_points_around<true>(fpos, r, [&](float& ref, double) {
+                ref = std::min(ref, -atom.occ);
+            }, false);
+          }
+        }
+      }
+    }
+
+    // reduce starting mask
+    for (size_t i = 0; i < m.data.size(); ++i) {
+      if (m.data[i] < 0.0)
+        mask.data[i] += m.data[i];
+    }
+  }
+
+  // When we reduced the solvent mask by the atom occupancy each time
+  // we hit a grid point, we might end up with a value below zero (due
+  // to overlapping atoms) and need to limit it here.
+  for (float& d : mask.data)
+    d = std::max(d, 0.f);
+}
+
+/// All points close to a grid point which is currently set below a value are changed
+/// @param mask given mask
+/// @param r distance value (in A)
+/// @param value consider only grid points that are currently set to this value
+/// @param margin_value new value for grid points satsifying criteria
 template<typename T>
 void set_margin_around(Grid<T>& mask, double r, T value, T margin_value) {
   int du = (int) std::floor(r / mask.spacing[0]);
@@ -171,23 +240,30 @@ void set_margin_around(Grid<T>& mask, double r, T value, T margin_value) {
         }
       }
   if (stencil2.empty()) {
+    // r is small; it should be faster to go through masked points and
+    // for each one check if any point in given radius are unmasked.
     for (typename Grid<T>::Point p : mask)
-      if (*p.value != value) {
+      if (*p.value < value) {
         for (const auto& wvu : stencil1) {
           size_t idx = mask.index_near_zero(p.u + wvu[2], p.v + wvu[1], p.w + wvu[0]);
-          if (mask.data[idx] == value) {
+          if (mask.data[idx] >= value) {
             *p.value = margin_value;
             break;
           }
         }
       }
   } else {
+    // r is large; it should be faster to go through unmasked points
+    // and for each one check if it has masked near neighbors (stencil1).
+    // These neighbors get marked as margin. If all near neighbors are
+    // unmasked, we can skip further neighbors (they will be checked
+    // from other points).
     for (typename Grid<T>::Point p : mask) {
-      if (*p.value == value) {
+      if (*p.value >= value) {
         bool found = false;
         for (const auto& wvu : stencil1) {
           size_t idx = mask.index_near_zero(p.u + wvu[2], p.v + wvu[1], p.w + wvu[0]);
-          if (mask.data[idx] != value) {
+          if (mask.data[idx] < value) {
             mask.data[idx] = margin_value;
             found = true;
           }
@@ -195,7 +271,7 @@ void set_margin_around(Grid<T>& mask, double r, T value, T margin_value) {
         if (found)
           for (const auto& wvu : stencil2) {
             size_t idx = mask.index_near_zero(p.u + wvu[2], p.v + wvu[1], p.w + wvu[0]);
-            if (mask.data[idx] != value)
+            if (mask.data[idx] < value)
               mask.data[idx] = margin_value;
           }
       }
@@ -209,10 +285,12 @@ struct SolventMasker {
   AtomicRadiiSet atomic_radii_set;
   bool ignore_hydrogen;
   bool ignore_zero_occupancy_atoms;
+  bool use_atom_occupancy = false;
   double rprobe;
   double rshrink;
   double island_min_volume;
   double constant_r;
+  double requested_spacing = 0.;
 
   SolventMasker(AtomicRadiiSet choice, double constant_r_=0.) {
     set_radii(choice, constant_r_);
@@ -231,7 +309,7 @@ struct SolventMasker {
         island_min_volume = 0.;
         break;
       case AtomicRadiiSet::Cctbx:
-        rprobe = 1.11;
+        rprobe = 1.1;
         rshrink = 0.9;
         island_min_volume = 0.;
         break;
@@ -248,19 +326,34 @@ struct SolventMasker {
     }
   }
 
+  /// fill whole grid with 1
   template<typename T> void clear(Grid<T>& grid) const { grid.fill((T)1); }
 
+  /// set grid points around atoms to 0
   template<typename T> void mask_points(Grid<T>& grid, const Model& model) const {
-    if (atomic_radii_set == AtomicRadiiSet::Constant)
-      mask_points_in_constant_radius(grid, model, constant_r + rprobe, (T)0,
-                                     ignore_hydrogen, ignore_zero_occupancy_atoms);
-    else
-      mask_points_in_varied_radius(grid, model, atomic_radii_set, rprobe, (T)0,
-                                   ignore_hydrogen, ignore_zero_occupancy_atoms);
+    mask_points_in_radius(grid, model, atomic_radii_set, constant_r + rprobe, (T)0,
+                          ignore_hydrogen, ignore_zero_occupancy_atoms);
   }
 
+  void mask_points(Grid<float>& grid, const Model& model) const {
+    if (use_atom_occupancy)
+      mask_points_using_occupancy(grid, model, atomic_radii_set, constant_r + rprobe,
+                                  ignore_hydrogen, ignore_zero_occupancy_atoms);
+    else
+      mask_points<float>(grid, model);
+  }
+
+  /// Apply symmetry to grid (to fill whole grid). If we are using an
+  /// integer/binary mask this is done by distributing all 0-value
+  /// grid points - while for a float map/mask we are setting each
+  /// grid point to the minimum value.
   template<typename T> void symmetrize(Grid<T>& grid) const {
-    grid.symmetrize([&](T a, T b) { return a == (T)0 || b == (T)0 ? (T)0 : (T)1; });
+    if (std::is_same<T, std::int8_t>::value) {
+      grid.symmetrize([&](T a, T b) { return a == (T)0 || b == (T)0 ? (T)0 : (T)1; });
+    }
+    else {
+      grid.symmetrize([&](T a, T b) { return a < b ? a : b; });
+    }
   }
 
   template<typename T> void shrink(Grid<T>& grid) const {
@@ -278,6 +371,7 @@ struct SolventMasker {
 
   // Removes small islands of Land=1 in the sea of 0. Uses flood fill.
   // cf. find_blobs_by_flood_fill()
+  // Currently doesn't work mask from mask_points_using_occupancy().
   template<typename T> int remove_islands(Grid<T>& grid) const {
     if (island_min_volume <= 0)
       return 0;
@@ -331,20 +425,6 @@ struct SolventMasker {
 #endif
 };
 
-// TODO: add argument Box<Fractional> src_extent
-template<typename T>
-void interpolate_grid(Grid<T>& dest, const Grid<T>& src, const Transform& tr, int order=2) {
-  FTransform frac_tr = src.unit_cell.frac.combine(tr).combine(dest.unit_cell.orth);
-  size_t idx = 0;
-  for (int w = 0; w != dest.nw; ++w)
-    for (int v = 0; v != dest.nv; ++v)
-      for (int u = 0; u != dest.nu; ++u, ++idx) {
-        Fractional dest_fr = dest.get_fractional(u, v, w);
-        Fractional src_fr = frac_tr.apply(dest_fr);
-        dest.data[idx] = src.interpolate(src_fr, order);
-      }
-}
-
 struct NodeInfo {
   double dist_sq;  // distance from the nearest atom
   bool found = false;  // the mask flag
@@ -366,18 +446,19 @@ inline void mask_with_node_info(Grid<NodeInfo>& mask, const Model& model, double
     for (const Residue& res : chain.residues)
       for (const Atom& atom : res.atoms) {
         Fractional frac0 = mask.unit_cell.fractionalize(atom.pos);
-        mask.template do_use_points_in_box<true>(frac0, du, dv, dw,
-                      [&](NodeInfo& ni, const Position& delta, int u, int v, int w) {
-                        double d2 = delta.length_sq();
-                        if (d2 < ni.dist_sq) {
-                          ni.dist_sq = d2;
-                          ni.found = true;
-                          //ni.elem = atom.element;
-                          ni.u = u;
-                          ni.v = v;
-                          ni.w = w;
-                        }
-                      });
+        mask.template do_use_points_in_box<true>(
+            frac0, du, dv, dw,
+            [&](NodeInfo& ni, double d2, const Position&, int u, int v, int w) {
+              if (d2 < ni.dist_sq) {
+                ni.dist_sq = d2;
+                ni.found = true;
+                //ni.elem = atom.element;
+                ni.u = u;
+                ni.v = v;
+                ni.w = w;
+              }
+            },
+            radius);
       }
 }
 
@@ -408,7 +489,7 @@ template<typename T>
 void interpolate_grid_of_aligned_model2(Grid<T>& dest, const Grid<T>& src,
                                         const Transform& tr,
                                         const Model& dest_model, double radius,
-                                        int order=2) {
+                                        int order=1) {
   Grid<NodeInfo> mask;
   mask.copy_metadata_from(dest);
   mask_with_node_info(mask, dest_model, radius);
@@ -420,7 +501,7 @@ void interpolate_grid_of_aligned_model2(Grid<T>& dest, const Grid<T>& src,
     if (ni.found) {
       Fractional dest_fr = dest.get_fractional(ni.u, ni.v, ni.w);
       Fractional src_fr = frac_tr.apply(dest_fr);
-      dest.data[idx] = src.interpolate(src_fr, order);
+      dest.data[idx] = src.interpolate_value(src_fr, order);
     }
   }
 }
@@ -441,14 +522,14 @@ void add_soft_edge_to_mask(Grid<T>& grid, double width) {
         if (grid.data[idx] >= 1e-3) continue;
         double min_d2 = width2 + 1;
         Fractional fctr = grid.get_fractional(u, v, w);
-        grid.template use_points_in_box<true>(fctr, du, dv, dw,
-                          [&](T& point, const Position& delta, int, int, int) {
-                            if (point > 0.999) {
-                              double d2 = delta.length_sq();
-                              if (d2 < min_d2)
-                                min_d2 = d2;
-                            }
-                          });
+        grid.template use_points_in_box<true>(
+            fctr, du, dv, dw,
+            [&](T& point, double d2, const Position&, int, int, int) {
+              if (point > 0.999) {
+                if (d2 < min_d2)
+                min_d2 = d2;
+              }
+            });
         if (min_d2 < width2)
           grid.data[idx] = T(0.5 + 0.5 * std::cos(pi() * std::sqrt(min_d2) / width));
       }
