@@ -29,7 +29,7 @@ const option::Descriptor MapUsage[] = {
   { Diff, 0, "d", "diff", Arg::None,
     "  -d, --diff  \tUse difference map coefficients." },
   { Section, 0, "", "section", Arg::Required,
-    "  --section=NAME  \tMTZ dataset name or CIF block name" },
+    "  --section=NAME  \tMTZ dataset name or CIF block name." },
   { FLabel, 0, "f", "", Arg::Required,
     "  -f COLUMN  \tF column (MTZ label or mmCIF tag)." },
   { PhLabel, 0, "p", "", Arg::Required,
@@ -37,13 +37,13 @@ const option::Descriptor MapUsage[] = {
   { WeightLabel, 0, "", "weight", Arg::Required,
     "  --weight=COLUMN  \t(normally not needed) weighting for F." },
   { GridDims, 0, "g", "grid", Arg::Int3,
-    "  -g, --grid=NX,NY,NZ  \tGrid size (user-specified minimum)." },
+    "  -g, --grid=NX,NY,NZ  \tMinimum grid size." },
   { ExactDims, 0, "", "exact", Arg::None,
-    "  --exact  \tUse the exact grid size specified by --grid." },
+    "  --exact  \tUse the exact grid size specified with --grid." },
   { Sample, 0, "s", "sample", Arg::Float,
-    "  -s, --sample=NUMBER  \tSet spacing to d_min/NUMBER (3 is usual)." },
+    "  -s, --sample=NUMBER  \tSet spacing to d_min/NUMBER (3 is common)." },
   { AxesZyx, 0, "", "zyx", Arg::None,
-    "  --zyx  \tOutput axes Z Y X as fast, medium, slow (default is X Y Z)." },
+    "  --zyx  \tInvert axis order in output map: Z=fast and X=slow." },
   { GridQuery, 0, "G", "", Arg::None,
     "  -G  \tPrint size of the grid that would be used and exit." },
   { TimingFft, 0, "", "timing", Arg::None,
@@ -51,7 +51,9 @@ const option::Descriptor MapUsage[] = {
 };
 
 
-static std::array<const Mtz::Column*, 2>
+namespace {
+
+std::array<const Mtz::Column*, 2>
 get_mtz_map_columns(const Mtz& mtz, const char* section, bool diff_map,
                     const char* f_label, const char* phi_label) {
   const Mtz::Column* f_col = nullptr;
@@ -90,7 +92,7 @@ get_mtz_map_columns(const Mtz& mtz, const char* section, bool diff_map,
   return {{f_col, phi_col}};
 }
 
-static const Mtz::Column&
+const Mtz::Column&
 get_mtz_column(const Mtz& mtz, const char* section, const char* label) {
   const Mtz::Dataset* ds = nullptr;
   if (section)
@@ -117,6 +119,8 @@ void adjust_size(const DataProxy& data, std::array<int, 3>& size,
   }
 }
 
+} // anonymous namespace
+
 gemmi::Grid<float>
 read_sf_and_fft_to_map(const char* input_path,
                        const std::vector<option::Option>& options,
