@@ -1,10 +1,13 @@
 // Copyright 2017-2023 Global Phasing Ltd.
 
 #include <gemmi/mmcif.hpp>   // for string_to_int
+#include <array>
 #include <unordered_map>
 #include <gemmi/mmcif_impl.hpp> // for set_cell_from_mmcif
 #include <gemmi/atox.hpp>    // for string_to_int
 #include <gemmi/enumstr.hpp> // for entity_type_from_string, polymer_type_from_string
+#include <gemmi/numb.hpp>    // for as_number
+#include <gemmi/polyheur.hpp>  // for restore_full_ccd_codes
 
 namespace gemmi {
 
@@ -23,19 +26,23 @@ void copy_string(const cif::Table::Row& row, int n, std::string& dest) {
     dest = cif::as_string(row[n]);
 }
 
+template<typename T>
+SMat33<T> get_smat33(cif::Table::Row& row, int n) {
+  return SMat33<T>{(T) cif::as_number(row[n+0]),
+                   (T) cif::as_number(row[n+1]),
+                   (T) cif::as_number(row[n+2]),
+                   (T) cif::as_number(row[n+3]),
+                   (T) cif::as_number(row[n+4]),
+                   (T) cif::as_number(row[n+5])};
+}
+
 std::unordered_map<std::string, SMat33<float>> get_anisotropic_u(cif::Block& block) {
   cif::Table aniso_tab = block.find("_atom_site_anisotrop.",
                                     {"id", "U[1][1]", "U[2][2]", "U[3][3]",
                                      "U[1][2]", "U[1][3]", "U[2][3]"});
   std::unordered_map<std::string, SMat33<float>> aniso_map;
   for (auto ani : aniso_tab)
-    aniso_map.emplace(ani[0], SMat33<float>{
-                                (float) cif::as_number(ani[1]),
-                                (float) cif::as_number(ani[2]),
-                                (float) cif::as_number(ani[3]),
-                                (float) cif::as_number(ani[4]),
-                                (float) cif::as_number(ani[5]),
-                                (float) cif::as_number(ani[6])});
+    aniso_map.emplace(ani[0], get_smat33<float>(ani, 1));
   return aniso_map;
 }
 
@@ -203,7 +210,24 @@ void read_connectivity(cif::Block& block, Structure& st) {
     copy_string(row, kLinkId, c.link_id);
     c.type = connection_type_from_string(row.str(kConnTypeId));
     if (row.has2(kSym1) && row.has2(kSym2)) {
-      c.asu = (row.str(kSym1) == row.str(kSym2) ? Asu::Same : Asu::Different);
+      std::string s1 = row.str(kSym1);
+      std::string s2 = row.str(kSym2);
+      if (s1 == s2) {
+        c.asu = Asu::Same;
+      } else {
+        c.asu = Asu::Different;
+        size_t sep1 = s2.find('_');
+        size_t sep2 = s2.find('_');
+        if (sep1 != std::string::npos && sep1 + 4 == s1.size() &&
+            sep2 != std::string::npos && sep2 + 4 == s2.size()) {
+          if (s1[0] == '1' && s1[1] == '_')  // symop1 is usually 1_555
+            c.reported_sym[0] = no_sign_atoi(s2.c_str());
+          else
+            c.reported_sym[0] = 99;
+          for (size_t i = 1; i <= 3; ++i)
+            c.reported_sym[i] = s2[sep2 + i] - s1[sep1 + i];
+        }
+      }
     }
     copy_double(row, kDistValue, c.reported_distance);
     for (int i = 0; i < 2; ++i) {
@@ -248,7 +272,7 @@ void read_prot_cis(cif::Block& block, Structure& st) {
                               "?pdbx_label_comp_id_2", "?pdbx_auth_comp_id_2", // 9-10
                               "?label_alt_id", "?pdbx_omega_angle"})) {        // 11-12
     CisPep cispep;
-    cispep.model_str = row.str(kModelNum);
+    cispep.model_num = cif::as_int(row[kModelNum], 0);
     cispep.partner_c.chain_name = row.str(kAuthAsymId);
     cispep.partner_c.res_id.seqid = make_seqid(row.str(kAuthSeqId), row.ptr_at(kInsCode));
     cispep.partner_c.res_id.name = cif::as_string(row.one_of(kAuthCompId, kLabelCompId));
@@ -256,7 +280,6 @@ void read_prot_cis(cif::Block& block, Structure& st) {
       cispep.partner_n.chain_name = row.str(kAuthAsymId2);
     if (row.has(kAuthSeqId2))
       cispep.partner_n.res_id.seqid = make_seqid(row.str(kAuthSeqId2), row.ptr_at(kInsCode2));
-    if (row.has(kLabelCompId2))
     cispep.partner_n.res_id.name = cif::as_string(row.one_of(kAuthCompId2, kLabelCompId2));
     if (row.has(kAltId))
       cispep.only_altloc = cif::as_char(row[kAltId], '\0');
@@ -478,6 +501,40 @@ DiffractionInfo* find_diffrn(Metadata& meta, const std::string& diffrn_id) {
   return nullptr;
 }
 
+// optimized Row::one_of(), use with care or it will crash a program
+struct RowAccess {
+  const std::string *val = nullptr;
+  const std::string *fallback = nullptr;
+  RowAccess(const cif::Table& tab, int n1, int n2) {
+    int pos1 = tab.positions.at(n1);
+    int pos2 = tab.positions.at(n2);
+    if (pos1 < 0) {
+      pos1 = pos2;
+      pos2 = -1;
+    }
+    const cif::Loop* loop = const_cast<cif::Table&>(tab).get_loop();
+    if (pos1 >= 0)
+      val = loop ? &loop->values[pos1] : &tab.bloc.items[pos1].pair[1];
+    if (pos2 >= 0)
+      fallback = loop ? &loop->values[pos2] : &tab.bloc.items[pos2].pair[1];
+  }
+  bool ok() const { return val != nullptr; }
+  const std::string& get(size_t gap) const {
+    const std::string& r = val[gap];
+    if (!cif::is_null(r) || fallback == nullptr)
+      return r;
+    return fallback[gap];
+  }
+};
+
+template<class T>
+T* get_by_id(std::vector<T>& vec, const std::string& id) {
+  for (T& item : vec)
+    if (item.id == id)
+      return &item;
+  return nullptr;
+}
+
 } // anonymous namespace
 
 Structure make_structure_from_block(const cif::Block& block_) {
@@ -523,7 +580,12 @@ Structure make_structure_from_block(const cif::Block& block_) {
                                           "?ls_d_res_high",           // 1
                                           "?ls_d_res_low",            // 2
                                           "?ls_percent_reflns_obs",   // 3
-                                          "?ls_number_reflns_obs"})) {
+                                          "?ls_number_reflns_obs",    // 4
+                                          "?ls_number_reflns_R_work", // 5
+                                          "?ls_number_reflns_R_free", // 6
+                                          "?ls_R_factor_obs",         // 7
+                                          "?ls_R_factor_R_work",      // 8
+                                          "?ls_R_factor_R_free"})) {
     st.meta.refinement.emplace_back();
     RefinementInfo& ref = st.meta.refinement.back();
     ref.id = row.str(0);
@@ -536,6 +598,57 @@ Structure make_structure_from_block(const cif::Block& block_) {
     copy_double(row, 2, ref.resolution_low);
     copy_double(row, 3, ref.completeness);
     copy_int(row, 4, ref.reflection_count);
+    copy_int(row, 5, ref.work_set_count);
+    copy_int(row, 6, ref.rfree_set_count);
+    copy_double(row, 7, ref.r_all);
+    copy_double(row, 8, ref.r_work);
+    copy_double(row, 9, ref.r_free);
+  }
+
+  for (auto row : block.find("_pdbx_refine_tls.", {
+        "id", "?pdbx_refine_id",
+        "T[1][1]", "T[2][2]", "T[3][3]", "T[1][2]", "T[1][3]", "T[2][3]",
+        "L[1][1]", "L[2][2]", "L[3][3]", "L[1][2]", "L[1][3]", "L[2][3]",
+        "S[1][1]", "S[1][2]", "S[1][3]",
+        "S[2][1]", "S[2][2]", "S[2][3]",
+        "S[3][1]", "S[3][2]", "S[3][3]",
+        "origin_x", "origin_y", "origin_z"})) {
+    if (st.meta.refinement.empty())
+      break;
+    RefinementInfo* ref = nullptr;
+    if (row.has(1))
+      ref = get_by_id(st.meta.refinement, row.str(1));
+    if (!ref)
+      ref = &st.meta.refinement[0];
+    ref->tls_groups.emplace_back();
+    TlsGroup& tls = ref->tls_groups.back();
+    tls.id = row.str(0);
+    tls.T = get_smat33<double>(row, 2);
+    tls.L = get_smat33<double>(row, 8);
+    for (int i = 0; i < 3; ++i)
+      for (int j = 0; j < 3; ++j)
+        tls.S[i][j] = cif::as_number(row[14+3*i+j]);
+    tls.origin.x = cif::as_number(row[23]);
+    tls.origin.y = cif::as_number(row[24]);
+    tls.origin.z = cif::as_number(row[25]);
+  }
+  for (auto row : block.find("_pdbx_refine_tls_group.", {
+        "refine_tls_id", "?beg_auth_asym_id", "?beg_auth_seq_id", "?beg_PDB_ins_code",
+        "?end_auth_seq_id", "?end_PDB_ins_code", "?selection_details"})) {
+    for (RefinementInfo& ref : st.meta.refinement)
+      if (TlsGroup* tls = get_by_id(ref.tls_groups, row.str(0))) {
+        tls->selections.emplace_back();
+        TlsGroup::Selection& sel = tls->selections.back();
+        if (row.has(1))
+          sel.chain = row.str(1);
+        if (row.has(2))
+          sel.res_begin = make_seqid(row.str(2), row.ptr_at(3));
+        if (row.has(4))
+          sel.res_end = make_seqid(row.str(4), row.ptr_at(5));
+        if (row.has(6))
+          sel.details = row.str(6);
+        break;
+      }
   }
 
   for (auto row : block.find("_exptl.", {"method", "?crystals_number"})) {
@@ -627,7 +740,10 @@ Structure make_structure_from_block(const cif::Block& block_) {
   for (auto row : block.find("_software.", {"name",
                                             "?classification",
                                             "?version",
-                                            "?date"})) {
+                                            "?date",
+                                            "?description",
+                                            "?contact_author",
+                                            "?contact_author_email"})) {
     st.meta.software.emplace_back();
     SoftwareItem& item = st.meta.software.back();
     item.name = row.str(0);
@@ -635,6 +751,9 @@ Structure make_structure_from_block(const cif::Block& block_) {
       item.classification = software_classification_from_string(row.str(1));
     copy_string(row, 2, item.version);
     copy_string(row, 3, item.date);
+    copy_string(row, 4, item.description);
+    copy_string(row, 5, item.contact_author);
+    copy_string(row, 6, item.contact_author_email);
   }
 
   std::vector<std::string> ncs_oper_tags = transform_tags("matrix", "vector");
@@ -698,7 +817,7 @@ Structure make_structure_from_block(const cif::Block& block_) {
                                       "occupancy",
                                       "B_iso_or_equiv",
                                       "?pdbx_formal_charge",
-                                      "auth_seq_id",
+                                      "?auth_seq_id",
                                       "?auth_comp_id",
                                       "?auth_asym_id",
                                       "?auth_atom_id",
@@ -708,36 +827,47 @@ Structure make_structure_from_block(const cif::Block& block_) {
                                       "?ccp4_deuterium_fraction",
                                      });
   if (atom_table.length() != 0) {
-    const int kAsymId = atom_table.first_of(kAuthAsymId, kLabelAsymId);
+    RowAccess asym_id(atom_table, kAuthAsymId, kLabelAsymId);
     // we use only one comp (residue) and one atom name
-    const int kCompId = atom_table.first_of(kAuthCompId, kLabelCompId);
-    const int kAtomId = atom_table.first_of(kAuthAtomId, kLabelAtomId);
-    if (!atom_table.has_column(kCompId))
+    RowAccess comp_id(atom_table, kAuthCompId, kLabelCompId);
+    RowAccess atom_id(atom_table, kAuthAtomId, kLabelAtomId);
+    RowAccess seq_id(atom_table, kAuthSeqId, kLabelSeqId);
+    if (!asym_id.ok())
+      fail("Neither _atom_site.label_asym_id nor auth_asym_id found");
+    if (!comp_id.ok())
       fail("Neither _atom_site.label_comp_id nor auth_comp_id found");
-    if (!atom_table.has_column(kAtomId))
+    if (!atom_id.ok())
       fail("Neither _atom_site.label_atom_id nor auth_atom_id found");
+    if (!seq_id.ok())
+      fail("Neither _atom_site.label_seq_id nor auth_seq_id found");
+    size_t loop_width = 0;
+    if (const cif::Loop* loop = atom_table.get_loop())
+      loop_width = loop->width();
 
     st.has_d_fraction = atom_table.has_column(kDeuterium);
 
     Model *model = nullptr;
     Chain *chain = nullptr;
     Residue *resi = nullptr;
-    if (atom_table.has_column(kModelNum))
-      model = &st.find_or_add_model(atom_table[0].str(kModelNum));
-    else
-      model = &st.find_or_add_model("1");
+    std::string model_num;
+    if (!atom_table.has_column(kModelNum)) {
+      st.models.emplace_back(1);
+      model = &st.models[0];
+    }
     for (auto row : atom_table) {
-      if (row.has(kModelNum) && row[kModelNum] != model->name) {
-        model = &st.find_or_add_model(row.str(kModelNum));
+      size_t gap = row.row_index * loop_width;
+      if (row.has(kModelNum) && row[kModelNum] != model_num) {
+        model_num = row[kModelNum];
+        model = &st.find_or_add_model(cif::as_int(model_num, 0));
         chain = nullptr;
       }
-      if (!chain || cif::as_string(row[kAsymId]) != chain->name) {
-        model->chains.emplace_back(cif::as_string(row[kAsymId]));
+      if (!chain || cif::as_string(asym_id.get(gap)) != chain->name) {
+        model->chains.emplace_back(cif::as_string(asym_id.get(gap)));
         chain = &model->chains.back();
         resi = nullptr;
       }
-      ResidueId rid = make_resid(cif::as_string(row[kCompId]),
-                                 cif::as_string(row[kAuthSeqId]),
+      ResidueId rid = make_resid(cif::as_string(comp_id.get(gap)),
+                                 cif::as_string(seq_id.get(gap)),
                                  row.has(kInsCode) ? &row[kInsCode] : nullptr);
       if (!resi || !resi->matches(rid)) {
         resi = chain->find_or_add_residue(rid);
@@ -759,7 +889,7 @@ Structure make_structure_from_block(const cif::Block& block_) {
         fail("Inconsistent sequence ID: " + resi->str() + " / " + rid.str());
       }
       Atom atom;
-      atom.name = cif::as_string(row[kAtomId]);
+      atom.name = cif::as_string(atom_id.get(gap));
       // altloc is always a single letter (not guaranteed by the mmCIF spec)
       atom.altloc = cif::as_char(row[kAltId], '\0');
       atom.charge = row.has2(kCharge) ? cif::as_int(row[kCharge]) : 0;
@@ -801,16 +931,21 @@ Structure make_structure_from_block(const cif::Block& block_) {
   }
 
   cif::Table polymer_types = block.find("_entity_poly.", {"entity_id", "type"});
-  for (auto row : block.find("_entity.", {"id", "type"})) {
+  for (auto row : block.find("_entity.", {"id", "?type"})) {
     Entity ent(row.str(0));
-    ent.entity_type = entity_type_from_string(row.str(1));
+    if (row.has(1))
+      ent.entity_type = entity_type_from_string(row.str(1));
     ent.polymer_type = PolymerType::Unknown;
     if (polymer_types.ok()) {
       try {
         std::string poly_type = polymer_types.find_row(ent.name).str(1);
+        if (ent.entity_type == EntityType::Unknown)
+          ent.entity_type = EntityType::Polymer;
         ent.polymer_type = polymer_type_from_string(poly_type);
       } catch (std::runtime_error&) {}
     }
+    // _entity_poly_seq is supposed to reflect heterogeneities in _atom_site.
+    ent.reflects_microhetero = true;
     st.entities.push_back(ent);
   }
 
@@ -822,7 +957,7 @@ Structure make_structure_from_block(const cif::Block& block_) {
       if (pos == (int) ent->full_sequence.size())
         ent->full_sequence.push_back(row.str(2));
       else if (pos >= 0 && pos < (int) ent->full_sequence.size())
-        ent->full_sequence[pos] += "," + row.str(2);
+        cat_to(ent->full_sequence[pos], ',', row.str(2));
     }
 
   cif::Table struct_ref = block.find("_struct_ref.",
@@ -895,7 +1030,62 @@ Structure make_structure_from_block(const cif::Block& block_) {
   st.assemblies = read_assemblies(block);
   read_sifts_unp(block, st);
 
+  cif::Table chem_comp_table = block.find("_chem_comp.", {"id", "three_letter_code"});
+  if (chem_comp_table.ok()) {
+    for (auto row : chem_comp_table) {
+      std::string alias = row.str(0);
+      std::string long_id = row.str(1);
+      if (alias[0] == '~' && long_id[0] != '~' && long_id[0] != '\0')
+        st.shortened_ccd_codes.emplace_back(long_id, alias);
+    }
+    restore_full_ccd_codes(st);
+  }
+
   return st;
 }
 
+
+Residue make_residue_from_chemcomp_block(const cif::Block& block, ChemCompModel kind) {
+  std::array<std::string, 3> xyz_tags;
+  switch (kind) {
+    case ChemCompModel::Xyz:
+      xyz_tags = {{"x", "y", "z"}};
+      break;
+    case ChemCompModel::Example:
+      xyz_tags = {{"model_Cartn_x", "model_Cartn_y", "model_Cartn_z"}};
+      break;
+    case ChemCompModel::Ideal:
+      xyz_tags = {{"pdbx_model_Cartn_x_ideal",
+                   "pdbx_model_Cartn_y_ideal",
+                   "pdbx_model_Cartn_z_ideal"}};
+      break;
+  }
+  Residue res;
+  res.seqid.num = 1;
+  cif::Column col =
+    const_cast<cif::Block&>(block).find_values("_chem_comp_atom.comp_id");
+  if (col && col.length() > 0)
+    res.name = col[0];
+  else
+    res.name = block.name.substr(starts_with(block.name, "comp_") ? 5 : 0);
+  cif::Table table = const_cast<cif::Block&>(block).find("_chem_comp_atom.",
+          {"atom_id", "type_symbol", "?charge",
+           xyz_tags[0], xyz_tags[1], xyz_tags[2]});
+  res.atoms.resize(table.length());
+  int n = 0;
+  for (auto row : table) {
+    Atom& atom = res.atoms[n++];
+    atom.name = row.str(0);
+    atom.element = Element(row.str(1));
+    if (row.has2(2))
+      // Charge is defined as integer, but some cif files in the wild have
+      // trailing '.000', so we read it as floating-point number.
+      atom.charge = (signed char) std::round(cif::as_number(row[2]));
+    atom.pos = Position(cif::as_number(row[3]),
+                        cif::as_number(row[4]),
+                        cif::as_number(row[5]));
+  }
+  return res;
+}
+
 } // namespace gemmi
