@@ -11,6 +11,10 @@ cif::Document read_cif_gz(const std::string& path) {
   return cif::read(MaybeGzipped(path));
 }
 
+bool check_cif_syntax_gz(const std::string& path, std::string* msg) {
+  return cif::check_syntax(MaybeGzipped(path), msg);
+}
+
 cif::Document read_mmjson_gz(const std::string& path) {
   return cif::read_mmjson(MaybeGzipped(path));
 }
@@ -19,8 +23,8 @@ CharArray read_into_buffer_gz(const std::string& path) {
   return read_into_buffer(MaybeGzipped(path));
 }
 
-cif::Document read_cif_from_buffer(const CharArray& buffer, const char* name) {
-  return cif::read_memory(buffer.data(), buffer.size(), name);
+cif::Document read_cif_from_memory(const char* data, size_t size, const char* name) {
+  return cif::read_memory(data, size, name);
 }
 
 cif::Document read_first_block_gz(const std::string& path, size_t limit) {
